package org.rzo.yacron4j;

import io.nettyx.util.Timeout;
import io.nettyx.util.Timer;
import io.nettyx.util.TimerTask;
import io.nettyx.util.internal.logging.InternalLogger;

import java.util.Date;
import java.util.concurrent.Callable;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.Delayed;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.quartzx.CronExpression;
import org.rzo.yacron4j.persist.PersistManager;

public class CronTaskImpl implements ScheduledFuture
{
	final Timer _timer;
	final CronExpression _exp;
	final AtomicLong _scheduledExecutionTime = new AtomicLong();
	final AtomicBoolean _terminated = new AtomicBoolean(false);
	final TimerTask _task;
	final AtomicReference<Timeout> _timeout = new AtomicReference<Timeout>();
	final TaskOptions _options;
	final AtomicInteger _concurrentExecs = new AtomicInteger(0);
	final CopyOnWriteArrayList<Thread> _executors = new CopyOnWriteArrayList<Thread>();
	final CopyOnWriteArrayList<Object> _results = new CopyOnWriteArrayList<Object>();
	final InternalLogger _logger;
	final AtomicBoolean _canceled = new AtomicBoolean(false);
	final Lock _lock = new ReentrantLock();
	final Condition _resultReady = _lock.newCondition();
	final AtomicReference<Date> _next = new AtomicReference<Date>();
	final String _id;
	final boolean _debug;
	final PersistManager _persistManager;
	final boolean _persist;

	class CronTimerTask implements TimerTask
	{

		protected final Runnable _runnable;
		protected final Callable _callable;

		public CronTimerTask(Runnable task)
		{
			_runnable = task;
			_callable = null;
		}

		public CronTimerTask(Callable task)
		{
			_callable = task;
			_runnable = null;
		}

		private void runTask()
		{
			while (true)
				try
				{
					_executors.add(Thread.currentThread());
					if (_debug)
						_logger.info(_id + ": exec start");
					if (_runnable != null)
						_runnable.run();
					else
					{
						Object result = _callable.call();
						_results.add(result);
					}
					if (_debug)
						_logger.info(_id + ": exec end");
					return;
				}
				catch (Throwable ex)
				{
					if (ex instanceof InterruptedException)
					{
						_logger.error(_id
								+ ": task execution interrupted -> abort", ex);
						Thread.currentThread().interrupt();
						return;
					}
					if (_options.getExceptionResolution() == TaskOptions.EXCEPTION_CANCEL)
					{
						_logger.error(_id + " task execution error -> cancel",
								ex);
						cancel(false);
						return;
					}
					if (_options.getExceptionResolution() == TaskOptions.EXCEPTION_IGNORE)
					{
						_logger.error(_id + " task execution error -> ignore",
								ex);
						return;
					}
					// else rerun
					_logger.error(_id + " task execution error -> retry", ex);
				}
				finally
				{
					_executors.remove(Thread.currentThread());
					_lock.lock();
					try
					{
						_resultReady.signal();
					}
					finally
					{
						_lock.unlock();
					}
				}
		}

		public void run(Timeout paramTimeout) throws Exception
		{
			if (_terminated.get())
				return;
			Date next = _next.get();
			boolean lastExecution = next == null;
			if (lastExecution)
			{
				runTask();
				terminate();
				return;
			}
			// if we allow 2 runs to overlap
			else if (_options.isAllowOverlap())
			{
				// schedule next start
				_scheduledExecutionTime.set(next.getTime());
				scheduleNextExecution();
				// run the task only if we do not have too many concurrent tasks
				if (_concurrentExecs.get() < _options.getMaxConcurrent())
				{
					_concurrentExecs.incrementAndGet();
					runTask();
					_concurrentExecs.decrementAndGet();
				}
				else
					_logger.warn(_id + ": too many concurrent tasks #"
							+ _concurrentExecs.get());

			}
			else if (_options.isIgnoreOnOverlap())
			{
				// if we do not allow overlaps and will ignore an exec in case
				// of overlap
				// first execute the task to see how long it takes
				runTask();
				// then schedule next execution
				_scheduledExecutionTime.set(next.getTime());
				if (_scheduledExecutionTime.get() < System.currentTimeMillis())
				{
					Date date = _exp.getTimeAfter(new Date());
					if (date == null)
					{
						persistTerminated();
						_terminated.set(true);
						return;
					}
					_scheduledExecutionTime.set(date.getTime());
				}
				scheduleNextExecution();
			}
			else
			// do not allow overlap and do not ignore executions
			{
				// first execute the task to see how long it takes
				runTask();
				// then schedule next execution. if this is in the past it will
				// be executed immediately
				_scheduledExecutionTime.set(next.getTime());
				scheduleNextExecution();
			}

		}

	}

	protected CronTaskImpl(String id, Timer timer, CronExpression exp,
			final Runnable task, TaskOptions options, PersistManager persistManager)
	{
		_id = id;
		_persist = options.isPersist() && !_id.startsWith("noname-");
		_persistManager = persistManager;
		_debug = options.isDebug();
		_exp = exp;
		_timer = timer;
		_scheduledExecutionTime.set(_exp.getTimeAfter(new Date()).getTime());
		_options = options;
		_logger = options.getLogger();
		_task = new CronTimerTask(task);
	}

	protected CronTaskImpl(String id, Timer timer, CronExpression exp,
			final Callable task, TaskOptions options, PersistManager persistManager)
	{
		_id = id;
		_persist = options.isPersist() && !_id.startsWith("noname-");
		_persistManager = persistManager;
		_debug = options.isDebug();
		_exp = exp;
		_timer = timer;
		_options = options;
		_logger = options.getLogger();
		_task = new CronTimerTask(task);
	}

	protected boolean start()
	{
		// check if the task is executable
		Date date = _exp.getTimeAfter(new Date());
		if (date == null)
		{
			_terminated.set(true);
			_logger.warn(_id + ": no next task execution -> ignore");
			return false;
		}
		Long persistedSchedule = _persistManager.getSchedule(_id);
		if (persistedSchedule != null)
			_scheduledExecutionTime.set(persistedSchedule.longValue());
		else
			_scheduledExecutionTime.set(date.getTime());
		return scheduleNextExecution();
	}

	private synchronized boolean scheduleNextExecution()
	{
		if (_terminated.get())
			return false;
		persistSchedule(_scheduledExecutionTime.get());
		if (_debug)
			_logger.info(_id + ": scheduled next execution: "
					+ new Date(_scheduledExecutionTime.get()));
		_next.set(_exp.getTimeAfter(new Date(_scheduledExecutionTime.get())));
		long delay = _scheduledExecutionTime.get() - System.currentTimeMillis();
		if (delay < 0)
			delay = 0;
		if (!_timer.isStarted())
			return false;
		_timeout.set(_timer.newTimeout(_task, delay - 10, TimeUnit.MILLISECONDS));
		return true;
	}

	public long getDelay(TimeUnit timeUnit)
	{
		return TimeUnit.MILLISECONDS.convert(
				_scheduledExecutionTime.get() - System.currentTimeMillis(), timeUnit);
	}

	public synchronized boolean cancel(boolean mayInterruptIfRunning)
	{
		if (_debug)
			_logger.info(_id + ": cancel");
		persistTerminated();
		_canceled.set(true);
		_terminated.set(true);
		if (_timeout != null)
			return _timeout.get().cancel();
		if (mayInterruptIfRunning)
			for (Thread thread : _executors)
				thread.interrupt();
		return true;
	}

	private void persistTerminated()
	{
		if (_persist)
			_persistManager.terminated(_id);
	}
	
	private void persistSchedule(long schedule)
	{
		if (_persist)
			_persistManager.scheduled(_id, schedule);
	}

	public int compareTo(Delayed o)
	{
		long result = getDelay(TimeUnit.MILLISECONDS)
				- o.getDelay(TimeUnit.MILLISECONDS);
		if (result > 0)
			return 1;
		if (result < 0)
			return -1;
		return 0;
	}

	public boolean isCancelled()
	{
		return _canceled.get();
	}

	public boolean isDone()
	{
		return _terminated.get();
	}

	public Object get() throws InterruptedException, ExecutionException
	{
		if (_canceled.get())
			throw new RuntimeException(_id + " task cancelled");
		if (_terminated.get())
			return getResult();
		if (_results.size() > 0)
			return getResult();
		_lock.lock();
		try
		{
			_resultReady.await();
		}
		finally
		{
			_lock.unlock();
		}
		return getResult();
	}

	private synchronized Object getResult()
	{
		if (_canceled.get())
			throw new RuntimeException(_id + " task cancelled");

		if (_results.size() == 0)
			return null;
		else
		{
			return _results.remove(0);
		}
	}

	public Object get(long timeout, TimeUnit unit) throws InterruptedException,
			ExecutionException, TimeoutException
	{
		if (_canceled.get())
			throw new RuntimeException(_id + " task cancelled");
		if (_terminated.get())
			return getResult();
		if (_results.size() > 0)
			return getResult();
		_lock.lock();
		try
		{
			_resultReady.await(timeout, unit);
		}
		finally
		{
			_lock.unlock();
		}
		return getResult();
	}

	private void terminate()
	{
		if (_debug)
			_logger.info(_id + ": terminated");
		persistTerminated();
		_terminated.set(true);
		_timeout.set(null);
		_scheduledExecutionTime.set(-1);
		_lock.lock();
		_resultReady.signal();
		_lock.unlock();
	}

}
